from enum import Enum
from nmigen import *
from nmigen.hdl.rec import *
from nmigen.asserts import *
from nmigen.cli import main




class ArbitraryWidthMemoryLayout(Layout):
    def __init__(self, *, data_width, address_width): # the * forces keyword args
        super().__init__([
            # DATA
            ("r_data",        unsigned(data_width)),      # OUTPUT
            ("w_data",        unsigned(data_width)),      # INPUT

            # ADDRESS
            ("r_addr",        unsigned(address_width)),   # INPUT
            ("w_addr",        unsigned(address_width)),   # INPUT


            # CONTROL
            ("fault",          1),

            # MODE SELECT
            ("write_enable",   1),                        # INPUT

            # FLOW CONTROL UPSTREAM
            ("valid_in",       1),                        # INPUT
            ("ready_out",      1),                        # OUTPUT

            # FLOW CONTROL DOWNSTREAM
            ("valid_out",      1),                        # OUTPUT
            ("ready_in",       1),                        # INPUT

        ])

class ArbitraryWidthMemoryBus(Record):
    def __init__(self, *, data_width, address_width):
        super().__init__(ArbitraryWidthMemoryLayout(data_width=data_width, address_width=address_width))



class ArbitraryWidthMemory(Elaboratable):
        def __init__(self, *, data_width, address_width):
            self.data_width = data_width
            self.address_width = address_width

            self.bus = ArbitraryWidthMemoryBus(data_width=data_width, address_width=data_width)

        def elaborate(self, platform):
            m = Module()


            return m



# This is non-synthesizable but is intended to provide a model for formal verification.

class GoldenArbitraryWidthMemory(Elaboratable):
        def __init__(self, *, data_width, address_width, sim_memory_size):
            self.data_width = data_width
            self.address_width = address_width
            self.sim_memory_size = sim_memory_size
            self.memory = Memory(width=data_width, depth=sim_memory_size, init=[0x40,0x41,0x42, 0x43, 0x44, 0x45,0x46,0x47])

            self.bus = ArbitraryWidthMemoryBus(data_width=data_width, address_width=address_width)

        def elaborate(self, platform):
            m = Module()

            write_ptr = Signal(range(self.sim_memory_size))
            read_ptr  = Signal(range(self.sim_memory_size))

            m.submodules.read_port  = read_port  = self.memory.read_port()
            m.submodules.write_port = write_port = self.memory.write_port()
            bus = self.bus

            with m.If(bus.valid_in == 1):
                with m.If(bus.write_enable == 1):
                    m.d.comb += write_port.en.eq(1)
                    m.d.comb += write_port.addr.eq(bus.w_addr)
                    m.d.comb += write_port.data.eq(bus.w_data)
                with m.Else():
                    #m.d.comb += read_port.en.eq(0)
                    m.d.comb += read_port.addr.eq(bus.r_addr)
                    m.d.comb += bus.r_data.eq(read_port.data)
                    m.d.sync += bus.valid_out.eq(1)


            return m



class DummyPlug(Elaboratable):

    #def __init__(self):



    def elaborate(self, platform):
        m = Module()

        m.submodules.AWMem = AWMem = GoldenArbitraryWidthMemory(data_width=7, address_width=8, sim_memory_size=8)
        counter = Signal(8)


        m.d.sync += counter.eq(counter+1)

        with m.If(counter == 4):
            m.d.comb += AWMem.bus.valid_in.eq(1)
        m.d.comb += AWMem.bus.write_enable.eq(0)
        m.d.comb += AWMem.bus.r_addr.eq(counter)
        #m.d.comb += AWMem.bus.w_addr.eq(counter)
        #m.d.comb += AWMem.bus.w_data.eq(counter)


        return m



if __name__ == '__main__':
    baka =DummyPlug()
    main(baka)
    #platform.build(DummyPlug())
