from enum import Enum
from nmigen import *
from nmigen.hdl.rec import *
from nmigen.asserts import *
from nmigen.cli import main
from nmigen.lib.fifo import *


class PipeStage(Elaboratable):
    def __init__(self, *, width, registered_ready):
        self.width = width
        self.registered_ready = registered_ready

        self.upstream_valid_in =    Signal(1)
        self.upstream_ready_out =   Signal(1)
        self.upstream_data_in =     Signal(self.width)

        self.downstream_valid_out = Signal(1)
        self.downstream_ready_in =  Signal(1)
        self.downstream_data_out =  Signal(self.width)

    def elaborate(self, platform):
        m = Module()

        # FORMAL PROPERTIES

        # UPSTREAM INTERFACE
        # Stable Ready for the upstream interface cannot be guarantted if we have a register in
        # the way (otherwise we'd need a combinatorial circuit which would defeat the point).
        if(self.registered_ready == False):
            with m.If((Past(self.upstream_ready_out) & (~Past(self.upstream_valid_in))) == 1):
                m.d.comb += Assert(self.upstream_ready_out == 1)

        # # Assumption of Stable data/valid on upstream interface is not needed for this proof:
        # with m.If((~Past(self.upstream_ready_out) & (Past(self.upstream_valid_in))) == 1):
        #     m.d.comb += Assume(Stable(self.upstream_data_in) & self.upstream_valid_in)

        # Cover interface transactions and stalls:
        m.d.comb += Cover(self.upstream_ready_out     & self.upstream_valid_in)
        m.d.comb += Cover((~self.upstream_ready_out)  & self.upstream_valid_in)


        # DOWNSTREAM INTERFACE

        # # Assumption of Stable Ready on downstream interface is not needed for this proof
        # with m.If((Past(self.downstream_ready_in) & (~Past(self.downstream_valid_out))) == 1):
        #     m.d.comb += Assume(self.downstream_ready_in == 1)

        with m.If(Past(ResetSignal()) == 0):
            with m.If((~Past(self.downstream_ready_in) & (Past(self.downstream_valid_out))) == 1):
                m.d.comb += Assert(Stable(self.downstream_data_out) & self.downstream_valid_out)


        # Cover interface transactions and stalls:
        m.d.comb += Cover(self.downstream_ready_in    & self.downstream_valid_out)
        m.d.comb += Cover((~self.downstream_ready_in) & self.downstream_valid_out)


        # IMPLEMENTATION

        if (self.registered_ready == False):
            m.d.comb += self.upstream_ready_out.eq(self.downstream_ready_in | (~self.downstream_valid_out))

            with m.If(self.upstream_ready_out == 1):
                m.d.sync +=  self.downstream_data_out.eq(self.upstream_data_in)
                m.d.sync += self.downstream_valid_out.eq(self.upstream_valid_in)

        else:
            pipe_reg = Signal(self.width)
            skid_reg = Signal(self.width)

            pipe_reg_valid = Signal(1)
            skid_reg_valid = Signal(1)

            m.d.comb += self.downstream_valid_out.eq( skid_reg_valid | pipe_reg_valid)
            m.d.comb +=   self.upstream_ready_out.eq(~skid_reg_valid)

            with m.If(skid_reg_valid == 1):
                m.d.comb += self.downstream_data_out.eq(skid_reg)
            with m.Else():
                m.d.comb += self.downstream_data_out.eq(pipe_reg)


            with m.If(self.upstream_ready_out == 1):
                m.d.sync +=       pipe_reg.eq(self.upstream_data_in)
                m.d.sync += pipe_reg_valid.eq(self.upstream_valid_in)
                with m.If(self.downstream_ready_in == 0):
                    m.d.sync +=       skid_reg.eq(pipe_reg)
                    m.d.sync += skid_reg_valid.eq(pipe_reg_valid)

            with m.If(self.downstream_ready_in == 1):
                m.d.sync += skid_reg_valid.eq(0)


            # FORMAL SECTION
            number_captured = Signal(3)

            upstream_transaction = Signal(1)
            downstream_transaction = Signal(1)

            m.d.comb += upstream_transaction.eq(self.upstream_ready_out & self.upstream_valid_in)
            m.d.comb += downstream_transaction.eq(self.downstream_ready_in & self.downstream_valid_out)

            m.d.sync += number_captured.eq(number_captured + upstream_transaction - downstream_transaction)

            with m.If(pipe_reg_valid == 0):
                with m.If(skid_reg_valid == 0):
                    m.d.comb += Assert(number_captured == 0)
                with m.Else():
                    m.d.comb += Assert(number_captured == 1)
            with m.Else():
                with m.If(skid_reg_valid == 0):
                    m.d.comb += Assert(number_captured == 1)
                with m.Else():
                    m.d.comb += Assert(number_captured == 2)


            m.d.comb += Assert((number_captured == 0) | (number_captured == 1) | (number_captured == 2))


            regular_sample = Signal(self.width)
            skid_sample    = Signal(self.width)


            with m.If(upstream_transaction == 1):
                with m.If(downstream_transaction == 1):
                    # Upstream transaction, downstream transaction
                    with m.If(number_captured == 2):
                        m.d.comb += Assert(self.downstream_data_out == skid_sample)
                    with m.Elif(number_captured == 1):
                        m.d.comb += Assert(self.downstream_data_out == regular_sample)
                        m.d.sync += regular_sample.eq(self.upstream_data_in)

                with m.Else():
                    # Upstream transaction, no downstream transaction
                    with m.If(number_captured == 0):
                        m.d.sync += regular_sample.eq(self.upstream_data_in)
                    with m.Elif(number_captured == 1):
                        m.d.sync += skid_sample.eq(regular_sample)
                        m.d.sync += regular_sample.eq(self.upstream_data_in)

            with m.Else():
                with m.If(downstream_transaction == 1):
                    # no upstream transaction, downstream transaction
                    with m.If(number_captured == 2):
                        m.d.comb += Assert(self.downstream_data_out == skid_sample)
                    with m.Elif(number_captured == 1):
                        m.d.comb += Assert(self.downstream_data_out == regular_sample)

                with m.Else():
                    # no upstream transaction, no downstream transaction
                    m.d.comb += Assert(1==1)




        return m




class DummyPlug(Elaboratable):
    def elaborate(self, platform):
        m = Module()
        #oldpipe = OldPipeStage(8)
        newpipe = PipeStage(width = 8, registered_ready = True)
        #newpipe = oldpipe
        #fifopipe = SyncFIFO(width = 8 , depth = 3,fwft=1)
        #egg = Trame(width = 8, alpha=fifopipe, beta=newpipe)
        m.submodules += newpipe

        m.d.comb += newpipe.upstream_valid_in.eq(AnySeq(1))
        m.d.comb += newpipe.downstream_ready_in.eq(AnySeq(1))
        m.d.comb += newpipe.upstream_data_in.eq(AnySeq(8))

        return m


if __name__ == '__main__':
    baka =DummyPlug()
    main(baka)